/*
Copyright (c) 2021 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifdef _MSC_VER
#  pragma warning(push)
#  pragma warning(disable : 4706) // suppress warning due to assignment within conditional expression
#  if _MSC_VER >= 1914
#    pragma warning(disable : 5045) // suppress information that MSVC will insert Spectre mitigation if /Qspectre switch specified
#  endif
#endif // _MSC_VER

#include "findinfile.h"
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#  include <share.h> // definition of _SH_DENYNO
#  define _ffind_fopen(_filename, _mode)         _fsopen(_filename, _mode, _SH_DENYNO) // fopen is marked both unsafe and deprecated for Windows
#  define _ffind_ftell(_file)                    _ftelli64(_file)                   // usually implementations on Windows x64 use data model LLP64, ...
#  define _ffind_fseek(_file, _offset, _origin)  _fseeki64(_file, _offset, _origin) // ... so we would still stuck with a 32 bit 'long' type without this branch
typedef int64_t _ffind_long_t;
#else // ! _WIN32
#  define _ffind_fopen(_filename, _mode)         fopen(_filename, _mode)
#  define _ffind_ftell(_file)                    ftell(_file)
#  define _ffind_fseek(_file, _offset, _origin)  fseek(_file, _offset, _origin)
typedef long _ffind_long_t;
#endif // ! _WIN32

// Help rounding upward to the next power of two that is not less than n. Relies on n > 0.
static inline size_t _ceil_pow_two(const long n)
{
    size_t c = (const size_t)n - 1U;
    c |= c >> 1U;
    c |= c >> 2U;
    c |= c >> 4U;
    c |= c >> 8U;
    c |= c >> 16U;
#if (ULONG_MAX > 4294967295UL) // most 64 bit platforms but Windows
    c |= c >> 32U;
#endif
    return c + 1U;
}

// Implements the Rabin-Karp search algorithm.
static FILE *_ffind_impl(FILE *file_stream, const void *needle, long needle_cnt)
{
  // BUFSIZ is likely the number of bytes used by the internal stream buffer and thus, it's the size we prefer to read at once.
  // If the length of the needle sequence is greater than BUFSIZ, we would get overlapped memory of the preceded bytes in our buffer,
  // and the position from where we copied these bytes. We round needle_cnt up to the next power of two in this case.
  const size_t chunk_max = (needle_cnt <= BUFSIZ) ? BUFSIZ : _ceil_pow_two(needle_cnt);

  // The total buffer size consists of preceded bytes where we copy the last needle_cnt bytes of the previous chunk (if any), and the chunk size in which we read new bytes from the file.
  unsigned char *const buffer = malloc((size_t)needle_cnt + chunk_max);
  if (!buffer)
    return NULL;

  unsigned char *const chunk_start = buffer + needle_cnt; // Position in the buffer which is the pointer where the newly read chunk begins.
  size_t read_cnt = fread(chunk_start, 1U, chunk_max, file_stream); // Read the first chunk and capture the length that have been read.
  if (read_cnt < (size_t)needle_cnt) // If the file content is shorter than needle it will be meaningless to begin searching.
  {
    free(buffer);
    return NULL;
  }

  const long hash_bitcnt = (long)(sizeof(uintmax_t) * CHAR_BIT); // size of the hashes in bits (constant expression at compile time)
  const int is_long_needle = (needle_cnt > hash_bitcnt); // 1 if needle is longer than the bit size of the hashes, 0 otherwise
  const uintmax_t eliminator = is_long_needle ? (uintmax_t)0U : (uintmax_t)1U << (unsigned)(needle_cnt - 1L); // factor used to remove bytes from chunk_hash
  uintmax_t chunk_hash = 0U, // rolling hash of the buffered bytes of the file
    needle_hash = 0U; // rolling hash of the bytes in needle
  const unsigned char *seq_cursor = chunk_start, // points to the buffered byte were the sequence begins, which will always be (chunk_cursor - needle_cnt) later on
    *chunk_cursor = is_long_needle ? chunk_start + (needle_cnt - hash_bitcnt) : chunk_start, // points to the buffered byte which we are about to process
    *chunk_end = chunk_start + read_cnt, // address which marks the end of the chunk read
    *const cpy_ptr = chunk_end - needle_cnt; // address from where to copy the last needle_cnt bytes if we have to read the next chunk

  // calculate both the hash of needle and the initial hash of the buffered bytes
  for (const unsigned char *const end = (const unsigned char*)needle + needle_cnt, *needle_cursor = is_long_needle ? end - hash_bitcnt : (const unsigned char*)needle; needle_cursor < end;)
  {
    needle_hash = (needle_hash << 1U) + *needle_cursor++;
    chunk_hash = (chunk_hash << 1U) + *chunk_cursor++;
  }

  for (;;) // Comparison and searching begins here.
  {
    // Same hashes indicate that we found the sequence. However, we still have to compare the whole sequence in this case.
    if (needle_hash == chunk_hash && memcmp(seq_cursor, needle, (size_t)needle_cnt) == 0)
    {
      // we need to move the stream cursor back to the begin of the found sequence
      if (_ffind_fseek(file_stream, (_ffind_long_t)(seq_cursor - chunk_end), SEEK_CUR) != 0)
        break;

      free(buffer);
      return file_stream;
    }

    // If all bytes of the chunk are processed, we need to read more data.
    if (chunk_cursor == chunk_end)
    {
      // we leave the function if we already reached the end of the file when we read the latest chunk
      if (read_cnt < chunk_max)
        break;

      // copy the last needle_cnt bytes to the begin of the buffer, also reset the cursor for the sequence
      seq_cursor = memcpy(buffer, cpy_ptr, (size_t)needle_cnt);
      // try to read the next chunk; leave the function if no further bytes have been read
      if ((read_cnt = fread(chunk_start, 1U, chunk_max, file_stream)) == 0U)
        break;

      chunk_cursor = chunk_start; // reset the cursor for the buffered bytes
      chunk_end = chunk_start + read_cnt; // update the end position of the chunk
    }

    // We update the hash of the buffered bytes.
    chunk_hash = ((chunk_hash - *seq_cursor++ * eliminator) << 1U) + *chunk_cursor++;
  }

  free(buffer);
  return NULL;
}

FILE *ffind(const char *file_name, FILE *file_stream, const void *needle, long needle_cnt)
{
  if (!needle || needle_cnt < 1L)
    return NULL;

  _ffind_long_t initial_pos = 0; // captures the current position of the stream cursor
  FILE *stream = NULL;
  if (file_stream)
  {
    // ensure we begin reading at the current position of the stream cursor (if the last action was a write operation)
    if (_ffind_fseek(file_stream, 0, SEEK_CUR) != 0)
      return NULL;

    stream = file_stream;
    initial_pos = _ffind_ftell(stream);
  }
  else if (!file_name || !(stream = _ffind_fopen(file_name, "r+b")))
    return NULL;

  FILE *ret = _ffind_impl(stream, needle, needle_cnt); // shared search implementation
  if (ret)
    return ret;

  if (file_stream)
    _ffind_fseek(stream, initial_pos, SEEK_SET); // reset the stream cursor if we didn't find the sequence
  else
    fclose(stream);

  return NULL;
}

#ifdef _WIN32
FILE *ffindw(const wchar_t *file_name, FILE *file_stream, const void *needle, long needle_cnt)
{
  if (!needle || needle_cnt < 1L)
    return NULL;

  _ffind_long_t initial_pos = 0; // captures the current position of the stream cursor
  FILE *stream = NULL;
  if (file_stream)
  {
    // ensure we begin reading at the current position of the stream cursor (if the last action was a write operation)
    if (_ffind_fseek(file_stream, 0, SEEK_CUR) != 0)
      return NULL;

    stream = file_stream;
    initial_pos = _ffind_ftell(stream);
  }
  else if (!file_name || !(stream = _wfsopen(file_name, L"r+b", _SH_DENYNO)))
    return NULL;

  FILE *ret = _ffind_impl(stream, needle, needle_cnt); // shared search implementation
  if (ret)
    return ret;

  if (file_stream)
    _ffind_fseek(stream, initial_pos, SEEK_SET); // reset the stream cursor if we didn't find the sequence
  else
    fclose(stream);

  return NULL;
}
#endif // _WIN32

#ifdef _MSC_VER
#  pragma warning(pop)
#endif // _MSC_VER
