#ifdef _MSC_VER
#  pragma warning(push)
// suppress meaningless information about inlining and that MSVC will insert Spectre mitigation if /Qspectre switch specified
#  pragma warning(disable : 4710 4711 5045)
#endif // _MSC_VER

#include "findinfile.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32 // fopen is marked both unsafe and deprecated for Windows
#  include <share.h> // definition of _SH_DENYNO
#  define FOPEN(_name, _mode) _fsopen(_name, _mode, _SH_DENYNO)
#else // ! _WIN32
#  define FOPEN(_name, _mode) fopen(_name, _mode)
#endif // ! _WIN32

#define FILE_NAME "foobar.txt"
#define HAYSTACK "Put a nice little secret into this file. Wanna know what? It's a secret."
#define NEEDLE "secret"
#define NEEDLE_LENGTH (sizeof(NEEDLE) - 1)
#define REPLACE "XXXXXX"
#define REPLACE_LENGTH (sizeof(REPLACE) - 1)
#define PREFACE 10000000

// Auxiliary function to create/overwrite a file using the haystack wherein we want to find the needle(s).
static void createfile(void)
{
  FILE *stream = FOPEN(FILE_NAME, "w");
  if (!stream)
    exit(1);

  void *buffer = malloc(PREFACE);
  if (!buffer)
  {
    fclose(stream);
    exit(1);
  }

  memset(buffer, 'A', PREFACE);
  fwrite(buffer, 1U, PREFACE, stream);
  free(buffer);
  fputs(HAYSTACK, stream);
  fclose(stream);
}

// Does finding the needle work correctly if the file is opened by the function?
static void test1(void)
{
  char buffer[128] = { 0 };
  createfile();
  // We pass the file name to the first parameter and NULL to the second parameter. This makes ffind open the file in "r+b" mode.
  // If the needle is found, the stream cursor of returned FILE* points to the related position. Also, the returned pointer must be closed using fclose.
  FILE *stream = ffind(FILE_NAME, NULL, NEEDLE, NEEDLE_LENGTH);
  if (!stream)
  {
    printf("%s%s\n", __func__, ": ERROR Needle not found.");
    return;
  }

  fwrite(REPLACE, 1U, REPLACE_LENGTH, stream); // Overwrite the needle.
  fseek(stream, PREFACE, SEEK_SET);
  size_t n = fread(buffer, 1U, 127U, stream);
  n = n;
  if (strcmp(buffer, "Put a nice little XXXXXX into this file. Wanna know what? It's a secret.") != 0) // This should be the new file content.
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR Wrong position.");
    return;
  }

  fclose(stream);
  printf("%s%s\n", __func__, ": OK.");
}

// Does finding the needle work correctly if searched in an already opened stream?
static void test2(void)
{
  char buffer[128] = { 0 };
  createfile();
  // It's not necessarily required to open the file in update mode.
  // However, for proper positioning of the stream cursor it must always be opened in binary mode.
  FILE *stream = FOPEN(FILE_NAME, "rb");
  if (!stream)
    exit(1);

  // We pass NULL to the first parameter and the FILE* to the second parameter. This makes ffind use the already opened stream.
  // If the function succeeds the returned pointer will be the same as the second parameter. Thus, it's sufficient to just check
  // if the function failed to find the needle. The returned pointer will be NULL in this case.
  if (!ffind(NULL, stream, NEEDLE, NEEDLE_LENGTH))
  {
    fclose(stream); // Even if the function fails, the stream is still open and valid. It must be closed.
    printf("%s%s\n", __func__, ": ERROR Needle not found.");
    return;
  }

  size_t n = fread(buffer, 1U, NEEDLE_LENGTH + 1, stream); // Check if we found the right position.
  n = n;
  if (strcmp(buffer, NEEDLE " ") != 0) // The space indicates that we actually found the first needle rather than the second.
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR Wrong position.");
    return;
  }

  fclose(stream);
  printf("%s%s\n", __func__, ": OK.");
}

// Is an updated stream cursor used as new starting position for subsequent searching?
static void test3(void)
{
  char buffer[128] = { 0 };
  createfile();
  FILE *stream = ffind(FILE_NAME, NULL, NEEDLE, NEEDLE_LENGTH); // Same as in test1.
  if (!stream)
  {
    printf("%s%s\n", __func__, ": ERROR First needle not found.");
    return;
  }

  size_t n = fread(buffer, 1U, NEEDLE_LENGTH + 1, stream); // Same as in test2. Furthermore, reading the first needle forwards the stream cursor.
  if (strcmp(buffer, NEEDLE " ") != 0)
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR First needle at wrong position.");
    return;
  }

  // Same as in test2, except that we start searching after the first needle.
  if (!ffind(NULL, stream, NEEDLE, NEEDLE_LENGTH))
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR Second needle not found.");
    return;
  }

  fwrite(REPLACE, 1U, REPLACE_LENGTH, stream); // Overwrite the second needle.
  fseek(stream, PREFACE, SEEK_SET);
  n = fread(buffer, 1U, 127U, stream);
  n = n;
  if (strcmp(buffer, "Put a nice little secret into this file. Wanna know what? It's a XXXXXX.") != 0) // This should be the new file content.
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR Second needle at wrong position.");
    return;
  }

  fclose(stream);
  printf("%s%s\n", __func__, ": OK.");
}

// Is the stream cursor moved back to the starting position if needle is not found?
static void test4(void)
{
  char buffer[128] = { 0 };
  createfile();
  FILE *stream = ffind(FILE_NAME, NULL, NEEDLE, NEEDLE_LENGTH); // Same as in test1.
  if (!stream)
  {
    printf("%s%s\n", __func__, ": ERROR Needle not found.");
    return;
  }

  // Same as in test3, except that we expect to fail finding the replace string.
  if (ffind(NULL, stream, REPLACE, REPLACE_LENGTH))
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR String should not have been found.");
    return;
  }

  size_t n = fread(buffer, 1U, 127U, stream);
  n = n;
  if (strcmp(buffer, "secret into this file. Wanna know what? It's a secret.") != 0) // This should be the remaining file content if the stream cursor is at the same position as before ffind was called.
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR Stream cursor changed position after failing to find the string.");
    return;
  }

  fclose(stream);
  printf("%s%s\n", __func__, ": OK.");
}

// How does it behave with needle having more bytes than the hash having bits, and even being longer than BUFSIZ?
static void test5(void)
{
  char buffer[BUFSIZ + BUFSIZ - 11] = { 0 };
  char needle[BUFSIZ + BUFSIZ - 11] = { 0 };
  char haystack[BUFSIZ + BUFSIZ + BUFSIZ + BUFSIZ + BUFSIZ] = { 0 };
  memset(needle, 'B', BUFSIZ + BUFSIZ - 11);
  memset(haystack, 'A', BUFSIZ + BUFSIZ + 11);
  memset(haystack + BUFSIZ + BUFSIZ + 11, 'B', BUFSIZ + BUFSIZ - 11);
  memset(haystack + BUFSIZ + BUFSIZ + BUFSIZ + BUFSIZ, 'A', BUFSIZ);
  FILE *stream = FOPEN(FILE_NAME, "w");
  if (!stream)
    exit(1);

  fwrite(haystack, 1U, BUFSIZ + BUFSIZ + BUFSIZ + BUFSIZ + BUFSIZ, stream);
  fclose(stream);

  stream = ffind(FILE_NAME, NULL, needle, BUFSIZ + BUFSIZ - 11);
  if (!stream)
  {
    printf("%s%s\n", __func__, ": ERROR Needle not found.");
    return;
  }

  size_t n = fread(buffer, 1U, BUFSIZ + BUFSIZ - 11, stream);
  n = n;
  if (memcmp(buffer, needle, BUFSIZ + BUFSIZ - 11) != 0) // The space indicates that we actually found the first needle rather than the second.
  {
    fclose(stream);
    printf("%s%s\n", __func__, ": ERROR Wrong position.");
    return;
  }

  fclose(stream);
  printf("%s%s\n", __func__, ": OK.");
}

int main(void)
{
  clock_t t1 = clock();
  test1();
  test2();
  test3();
  test4();
  test5();
  printf("%f s\n", ((double)clock() - (double)t1) / CLOCKS_PER_SEC);
  remove(FILE_NAME); // clean up
  return 0;
}

#ifdef _MSC_VER
#  pragma warning(pop)
#endif // _MSC_VER
